#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*
	Generate a Calabi-Yau surface geometry for POVRay
*/

typedef struct {
	double real,imag;
} COMPLEX;
typedef struct {
	double x,y,z;
} XYZ;
typedef struct {
	double r,g,b;
} COLOUR;

#define PI 3.141592653589793238462643
#define TWOPI 6.283185307179586476925287
#define N 30

COLOUR GetColour(double,double,double,int);
XYZ Eval(double,double,int,int,int,double);
COMPLEX Cpowd(COMPLEX,double);
COMPLEX Cmult(COMPLEX,COMPLEX);

int main(int argc,char **argv)
{
	int i,j,k1,k2,n=5;
	double a,b,da,db,theta=1/30.0;
	XYZ p[4];
	COLOUR c;
	char fname[64];
	FILE *fptr;

	da = 2 / (double)N;
	db = 0.5 * PI / (double)N;

	sprintf(fname,"calabiyau.inc");
	fptr = fopen(fname,"w");

	for (k1=0;k1<n;k1++) {
		for (k2=0;k2<n;k2++) {
			fprintf(fptr,"\n// k1:%d k2:%d\n\n",k1,k2);

			for (i=0;i<N;i++) {
				a = -1 + da * i;
	
				for (j=0;j<N;j++) {
					b = db * j;
	
					p[0] = Eval(a,b,k1,k2,n,theta);
               p[1] = Eval(a+da,b,k1,k2,n,theta);
               p[2] = Eval(a+da,b+db,k1,k2,n,theta);
               p[3] = Eval(a,b+db,k1,k2,n,theta);
	
					c = GetColour(a,-1.0,1.0,10);

					fprintf(fptr,"triangle {\n");
					fprintf(fptr," <%g,%g,%g>,\n",p[0].x,p[0].y,p[0].z);
               fprintf(fptr," <%g,%g,%g>,\n",p[1].x,p[1].y,p[1].z);
               fprintf(fptr," <%g,%g,%g> \n",p[2].x,p[2].y,p[2].z);
					fprintf(fptr," texture {\n");
					fprintf(fptr,"  pigment { color rgbt <%g,%g,%g,TRANSPARENCY> }\n",c.r,c.g,c.b);
					fprintf(fptr,"  finish { thefinish }\n");
					fprintf(fptr,"  normal { thenormal }\n");
					fprintf(fptr," }\n");
					fprintf(fptr,"}\n");
               fprintf(fptr,"triangle {\n");
               fprintf(fptr," <%g,%g,%g>,\n",p[2].x,p[2].y,p[2].z);
               fprintf(fptr," <%g,%g,%g>,\n",p[3].x,p[3].y,p[3].z);
               fprintf(fptr," <%g,%g,%g> \n",p[0].x,p[0].y,p[0].z);
               fprintf(fptr," texture {\n");
               fprintf(fptr,"  pigment { color rgbt <%g,%g,%g,TRANSPARENCY> }\n",c.r,c.g,c.b);
               fprintf(fptr,"  finish { thefinish }\n");
               fprintf(fptr,"  normal { thenormal }\n");
               fprintf(fptr," }\n");
               fprintf(fptr,"}\n");

				} // i
			} // j
		} // k2
	} // k1

	fclose(fptr);
	exit(0);
}

COLOUR GetColour(double v,double vmin,double vmax,int type)
{
   double dv,vmid;
   COLOUR c = {1.0,1.0,1.0};

   if (vmax < vmin) {
      dv = vmin;
      vmin = vmax;
      vmax = dv;
   }
   if (vmax - vmin < 0.000001) {
      vmin -= 1;
      vmax += 1;
   }
   if (v < vmin)
      v = vmin;
   if (v > vmax)
      v = vmax;
   dv = vmax - vmin;

   if (v < (vmin + 0.2 * dv)) {
      c.r = 0;
      c.g = 5 * (v - vmin) / dv;
      c.b = 1;
   } else if (v < (vmin + 0.4 * dv)) {
      c.r = 0;
      c.g = 1;
      c.b = 1 + 5 * (vmin + 0.2 * dv - v) / dv;
   } else if (v < (vmin + 0.6 * dv)) {
      c.r = 5 * (v - vmin - 0.4 * dv) / dv;
      c.g = 1;
      c.b = 0;
   } else if (v < (vmin + 0.8 * dv)) {
      c.r = 1;
      c.g = 1 - 5 * (v - vmin - 0.6 * dv) / dv;
      c.b = 0;
   } else {
      c.r = 1;
      c.g = 5 * (v - vmin - 0.8 * dv) / dv;
      c.b = 5 * (v - vmin - 0.8 * dv) / dv;
   }

   return(c);
}

XYZ Eval(double a,double b,int k1,int k2,int n,double theta)
{
   COMPLEX u1,u2;
	COMPLEX e1,e2;
   COMPLEX z1,z2;
	XYZ p;

   u1.real = 0.5 * (exp(a) * cos(b) + exp(-a) * cos(-b));
   u1.imag = 0.5 * (exp(a) * sin(b) + exp(-a) * sin(-b));
	u1 = Cpowd(u1,2.0/n);
   u2.real = 0.5 * (exp(a) * cos(b) - exp(-a) * cos(-b));
   u2.imag = 0.5 * (exp(a) * sin(b) - exp(-a) * sin(-b));
	u2 = Cpowd(u2,2.0/n);

	e1.real = cos(TWOPI * k1 / (double)n);
	e1.imag = sin(TWOPI * k1 / (double)n);
   e2.real = cos(TWOPI * k2 / (double)n);
   e2.imag = sin(TWOPI * k2 / (double)n);

   z1 = Cmult(e1,u1);
   z2 = Cmult(e2,u2);

   p.x = z1.real;
   p.y = z2.real;
   p.z = z1.imag * theta + z2.imag * (1-theta);

	return(p);
}

/*
   Cpowd(z,d) = z ^ d
   De Moivre's theorem
   Multiply angle by power and raise the modulus to the power
*/
COMPLEX Cpowd(COMPLEX z,double d)
{
   COMPLEX ztmp;
   double phi=0,r;

   r = z.real * z.real + z.imag * z.imag;
   r = exp(d * log(r) / 2);
   phi = d * atan2(z.imag,z.real);
   ztmp.real = r * cos(phi);
   ztmp.imag = r * sin(phi);

   return(ztmp);
}

/*
   Cmult(z1,z2) = z1 * z2
*/
COMPLEX Cmult(COMPLEX z1,COMPLEX z2)
{
    COMPLEX ztmp;

    ztmp.real = z1.real * z2.real - z1.imag * z2.imag;
    ztmp.imag = z1.real * z2.imag + z2.real * z1.imag;
    return(ztmp);
}

